%

%
% Copyright (c) 2015-2016,2019 LAAS/CNRS
% All rights reserved.
%
% Redistribution  and  use  in  source  and binary  forms,  with  or  without
% modification, are permitted provided that the following conditions are met:
%
%   1. Redistributions of  source  code must retain the  above copyright
%      notice and this list of conditions.
%   2. Redistributions in binary form must reproduce the above copyright
%      notice and  this list of  conditions in the  documentation and/or
%      other materials provided with the distribution.
%
% THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
% WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
% MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
% ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
% WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
% ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
% IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
%
%                                           Anthony Mallet on Tue Jun  2 2015
%
classdef client < handle
  properties (SetAccess = private)
    host
    session
    id
  end

  methods
    function self = client(host, session)
      % genomix.client - Create connections to a genomix server.
      %
      % c = genomix.client()
      %    Connect to a server on 'localhost' port 8080.
      % c = genomix.client('host:port')
      %    Connect to a server on 'host' and 'port'.
      %
      % genomix.client Methods:
      %    rpath - Add a new load path on the server.
      %    load - Load a component on the server.
      %    update - Process received events.
      %
      switch nargin
        case 0
          host = 'localhost:8080';
          session = '';
        case 1
          session = '';
      end

      self.host = host;
      self.session = session;
      try
        self.id = genomix_server('connect', self.host, self.session);
      catch e
        self.id = -1;
        rethrow(e);
      end
    end

    function delete(self)
      % delete - Disconnect an instance of a genomix.client.
      %
      rqstids = genomix.client.rqstids;
      for rid = keys(rqstids)
        r = rqstids(rid{1});
        if (~isvalid(r.rh))
          remove(rqstids, rid{1});
        elseif (r.rh.client == self)
          remove(rqstids, rid{1});
          delete(r.rh);
        end
      end

      genomix_server('disconnect', self.id);
    end


    % --- rpath ------------------------------------------------------------
    %
    function r = rpath(self, rpath)
      % rpath - Add a new load path on the server.
      %
      r = genomix_server('get', self.id, sprintf('/rpath/%s', rpath));
    end


    % --- load -------------------------------------------------------------
    %
    function r = load(self, client, varargin)
      % load - Load a component on the server.
      %
      % c = client.load(component) -
      %   Load 'component' in the server a return the corresponding
      %   genomix.component handle.
      %
      % c = client.load(component, '-h') -
      %   Return the available load options for the 'component'.
      %
      usage = 0;
      argv = {};
      for i = 1:length(varargin)
        if strcmp(varargin{i}, '-h')
          usage = 1;
        else
          argv = [argv varargin(i)];
        end
      end
      args.argv = strjoin(argv, ' ');

      if usage
        r = sprintf([ ...
           'Usage:\n\t client.load(''%s'', [-h], [client options])\n\n' ...
           'Client options:\n' ], client);
        opts = genomix_server('get', self.id, strcat('/usage/', client));
        for i = 1:length(opts)
            r = sprintf('%s  %-26s %s\n', r, opts{i}.option, opts{i}.descr);
        end
        return
      end

      module = genomix_server('get', self.id, ['/load/' client], args);
      meta = genomix_server('get', self.id,  ['/module/' module '/info']);
      r = genomix.component(self, module, meta);
    end


    % --- update -----------------------------------------------------------
    %
    function update(self)
      genomix_server('doevents', self.id, 0);
    end
  end

  methods (Static)
    % --- rqst -------------------------------------------------------------
    %
    function id = rqst(rqst, varargin)
      if ~isvalid(rqst.client)
        e.message = 'client was deleted';
        e.stack.file = '';
        e.stack.line = 0;
        e.stack.name = strcat(rqst.component, '::', rqst.name);
        error(e);
      end

      % process options
      flat = 0;
      oneway = 0;
      wait = 'done';
      timeout = -1;
      iarg = 1;
      narg = nargin - 1;

      while iarg <= narg
        if ~ischar(varargin{iarg}); break; end
        switch varargin{iarg}
          case { '-o', '-oneway' }
            oneway = 1;
          case { '-a', '-ack' }
            wait = 'sent';
          case { '-t', '-timeout' }
            timeout = varargin{iarg+1};
            iarg = iarg + 1;
          case { '-f', '-flat' }
            flat = 1;
          case '-args';
            id = rqst.meta;
            return;

          case '--'
            break;
          case '-h'
            id = sprintf([ 'Usage:\n\t %s.%s ' ...
                           '[-a|-ack] [-t|-timeout secs] [-f|-flat]\n\t\t' ...
                           '[-args] [-h] [--] input [callback]\n\n' ...
                           'Input dictionary:%s' ], ...
                         rqst.component, rqst.name, ...
                         genomix.client.meta2doc(rqst.meta.input, ...
                                                 rqst.meta.meta));
            return

          otherwise
            if varargin{iarg}(1) == '-'
              e.message = ['unknown option -- ' varargin{iarg}];
              e.stack.file = '';
              e.stack.line = 0;
              e.stack.name = strcat(rqst.component, '::', rqst.name);
              error(e);
            end
            break
        end
        iarg = iarg + 1;
      end

      if oneway
         args.oneway = 1;
      else
        % look for callback
        callback = [];
        if narg > 0 && isa(varargin{narg}, 'function_handle')
          callback = varargin{narg};
          narg = narg - 1;
          wait = 'sent';
        end
      end

      % create input argument
      switch narg - iarg + 1
        case 1
          if isstruct(varargin{iarg})
            % merge defaults
            for k = fieldnames(rqst.defarg)'
              if ~isfield(varargin{iarg}, k{1})
                varargin{iarg}.(k{1}) = rqst.defarg.(k{1});
              end
            end
            args.input.data = varargin{iarg};
          else
            args.input.data = ...
            genomix.client.arg2input(varargin(iarg:narg), rqst.meta.input);
          end

        case 0
          if flat; error('-f requires arguments'); end
          args.input.data = ...
          genomix.client.iscanargs(rqst.meta.input, rqst.meta.meta);

        otherwise
          args.input.data = ...
          genomix.client.arg2input(varargin(iarg:narg), rqst.meta.input);
      end
      args.input.format = rqst.meta.input;

      % send remotely
      try
        i = genomix_server('get', rqst.client.id, ...
                           ['/module/' rqst.component '/send/' rqst.name], ...
                           args);
      catch ME
        e.message = ME.message;
        e.stack.file = '';
        e.stack.line = 0;
        e.stack.name = strcat(rqst.component, '::', rqst.name);
        error(e);
      end

      if oneway; id = {}; return; end;

      % register request
      id = genomix.request(rqst.client, rqst.component, rqst.name, i);
      r = struct();
      r.rh = id;
      r.callback = callback;
      r.status = 'none';
      rqstids = genomix.client.rqstids;
      rqstids(sprintf('%s::%s', rqst.component, i)) = r;

      % wait for completion
      id.wait(timeout, wait);
    end


    % --- port -------------------------------------------------------------
    %
    function data = port(port, varargin)
      if ~isvalid(port.client)
        e.message = 'client was deleted';
        e.stack.file = '';
        e.stack.line = 0;
        e.stack.name = strcat(port.component, '::', port.name);
        error(e);
      end

      if port.meta.simple
        try
          data = genomix_server('get', port.client.id, ...
                                ['/module/' port.component '/read/' port.name]);
        catch ME
          e.message = ME.message;
          e.stack.file = '';
          e.stack.line = 0;
          e.stack.name = strcat(port.component, '::', port.name);
          error(e);
        end
      elseif ~port.meta.simple
        if nargin ~= 2
          e.message = 'missing port name argument';
          e.stack.file = '';
          e.stack.line = 0;
          e.stack.name = strcat(port.component, '::', port.name);
          error(e);
        end
        try
          data = genomix_server('get', port.client.id, ...
                                ['/module/' port.component '/read/' ...
                                            port.name '/' varargin{1}]);
        catch ME
          e.message = ME.message;
          e.stack.file = '';
          e.stack.line = 0;
          e.stack.name = strcat(port.component, '::', port.name);
          error(e);
        end
      end
    end


    % --- doevents --------------------------------------------------
    %
    function doevents(data)
      if ~isfield(data, 'id'); return; end

      rqstids = genomix.client.rqstids;
      if rqstids.isKey(data.id)
        r = rqstids(data.id);
        for p = fieldnames(data)'
          r.(p{1}) = data.(p{1});
          switch p{1}
            case 'status'
              r.rh.status = data.status;
            case 'output'
              r.rh.result = data.output;
            case 'errmsg'
              r.rh.exception = data.errmsg;
          end
        end

        if ~isempty(r.callback)
          cb = r.callback;
          cb(r);
        end

        if isfield(data, 'status')
          if strcmp(data.status, 'done') || strcmp(data.status, 'error')
            remove(rqstids, data.id);
            if isempty(rqstids.keys); munlock; end
          end
        end
      end
    end

    function val = rqstids()
      persistent pval;
      if isempty(pval)
        pval = containers.Map;
      end
      val = pval;
    end
  end

  % --- save and load ------------------------------------------------------
  %
  % The only relevant info is the host
  %
  methods
    function s = saveobj(obj)
      s = struct('host', obj.host);
    end
  end

  methods (Static)
    function obj = loadobj(s)
      if isstruct(s)
        obj = genomix.client(s.host);
      else
        obj = s;
      end
    end
  end


  methods (Static, Access = private)

    % --- iscanarg ---------------------------------------------------------
    %
    function input = iscanarg(kind, prompt, defval)
      if ~isempty(defval)
        if isnumeric(defval)
          prompt = sprintf('%s (%g)', prompt, defval);
        else
          prompt = sprintf('%s (%s)', prompt, defval);
        end
      end

      while 1
        p = sprintf('input(''%s > '')', prompt);
        input = evalin('base', p);

        % user might want default value or interrupt
        if isempty(input); input = defval; end
        if ischar(input) && strcmp(input, '...')
          input = {};
          return
        end

        % check input validity
        try
          switch kind{1}
            case 'boolean'
              if input; input = true; else input = false; end
            case 'octet'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < -128 || input > 127
                error('argument out of range, expecting %s', kind{1});
              end
            case 'short'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < -32768 || input > 32767
                error('argument out of range, expecting %s', kind{1});
              end
            case 'unsigned short'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < 0 || input > 65535
                error('argument out of range, expecting %s', kind{1});
              end
            case 'long'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < -2^31 || input > 2^31-1
                error('argument out of range, expecting %s', kind{1});
              end
            case 'unsigned long'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < 0 || input > 2^32-1
                error('argument out of range, expecting %s', kind{1});
              end
            case 'long long'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < -2^63 || input > 2^63-1
                error('argument out of range, expecting %s', kind{1});
              end
            case 'unsigned long long'
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
              if input < 0 || input > 2^64-1
                error('argument out of range, expecting %s', kind{1});
              end
            case {'float' 'double'}
              if ~isnumeric(input)
                error('expecting %s', kind{1});
              end
            case 'char'
              if ~ischar(input)
                error('expecting %s', kind{1});
              end
              input = input{1};
            case 'string'
              if ~ischar(input)
                error('expecting %s', kind{1});
              end
              if length(kind) > 1
                 if length(input) > kind{2}
                   error('string too long');
                 end
              end
            case 'enum'
              if find(ismember(kind{2}, input))
              else
                error('expecting %s', strjoin(kind{2}, ', '));
              end
          end
        catch e
          disp(e.message);
          continue;
        end

        break;
      end
    end


    % --- iscandict --------------------------------------------------------
    %
    function arg = iscandict(indent, name, format, meta)
      prompt = sprintf('%s %s %s', indent, format.kind, name);
      if isfield(format, 'length')
        prompt = sprintf('%s[%d]', prompt, format.length);
      end
      if isfield(meta, 'doc')
        prompt = sprintf('%s: %s', prompt, meta.doc);
      end
      if isfield(meta, 'items'); items = meta.items; else items = struct(); end
      if isfield(meta, 'default')
        default = meta.default;
      else
        default = [];
      end

      switch format.kind
        case {'struct' 'union'}
          disp(prompt);
          arg = struct();
          for p = fieldnames(format.type)'
            if isfield(items, p{1}); m = items.(p{1}); else m = struct(); end

            a = genomix.client.iscandict(sprintf('%s| ', indent), ...
                                         sprintf('%s.%s', name, p{1}), ...
                                         format.type.(p{1}), m);
            if iscell(a) && isempty(a); break; end
            arg.(p{1}) = a;
          end

        case {'array' 'sequence'}
          disp(prompt);
          if isfield(format, 'length')
            len = format.length;
          else
            len = genomix.client.iscanarg({'unsigned long'}, ...
                                          sprintf('%s length', prompt), 0);
            if isempty(len); arg = {}; return; end
          end

          arg = cell(1, len);
          for i = 1:len
            if iscell(items) && length(items) >= i
              m = items{i};
            else
              m = struct();
            end

            a = genomix.client.iscandict(sprintf('%s| ', indent), ...
                                         sprintf('%s[%d]', name, i), ...
                                         format.type, m);
            if iscell(a) && isempty(a); break; end
            arg{i} = a;
          end

        case 'optional'
          disp(prompt);
          pres = genomix.client.iscanarg({'boolean'}, ...
                                         sprintf('%s| present', indent), 1);
          if isempty(pres) || ~pres; arg = {}; return; end
          m = struct('items', items, 'default', default);
          a = genomix.client.iscandict(sprintf('%s| ', indent), name, ...
                                       format.type, m);
          if isempty(a); arg = {}; return; end
          arg = {a};

        otherwise
          kind = { format.kind };
          if isfield(format, 'length')
            kind(end+1) = { format.length };
          end
          if isfield(format, 'values')
            kind = [ kind format.values() ];
          end
          arg = genomix.client.iscanarg(kind, prompt, default);
      end
    end


    % --- iscanargs --------------------------------------------------------
    %
    function arg = iscanargs(input, meta)
      arg = struct();
      for p = fieldnames(input)'
        if isfield(meta, p{1}); m = meta.(p{1}); else m = struct(); end
        a = genomix.client.iscandict('', p{1}, input.(p{1}), m);
        if isempty(a); break; end
        arg.(p{1}) = a;
      end
    end


    % --- arg2input -------------------------------------------------
    %
    function args = arg2input(data, format)
      % create argument dictionary
      args = struct();
      for p = fieldnames(format)'
        if isempty(data); break; end
        [args.(p{1}), data] = genomix.client.args2dict(format.(p{1}), data);
      end

      if (~isempty(data))
        error('too many arguments: %d extra', length(data));
      end
    end


    % --- args2dict ---------------------------------------------------------
    %
    function [arg, args] = args2dict(format, args)
      switch format.kind
        case {'struct', 'union'}
          arg = struct();
          for p = fieldnames(format.type)'
            if isempty(args); return; end
            [arg.(p{1}), args] = ...
            genomix.client.args2dict(format.type.(p{1}), args);
          end

        case {'array', 'sequence'}
          if isempty(args); arg = {}; return; end
          a = args{1}; args(1) = [];
          arg = cell(1, length(a));
          for i = 1:length(a)
            [v, a] = genomix.client.args2dict(format.type, a);
            arg{i} = v;
          end

        case 'optional'
          if ~isempty(args{1})
            [arg, args] = genomix.client.args2dict(format.type, args);
          else
            arg = args{1}; args(1) = [];
          end

        otherwise
          if isempty(args)
            arg = [];
          elseif iscell(args)
            arg = args{1}; args(1) = [];
          else
            arg = args; args = {};
          end
      end
    end


    % --- meta2doc ---------------------------------------------------------
    %
    function doc = meta2doc(input, meta)
      doc = '';
      for p = fieldnames(input)'
        if isfield(meta, p{1}); m = meta.(p{1}); else m = struct(); end
        doc = [doc genomix.client.dict2doc('', p{1}, input.(p{1}), m)];
      end
    end


    % --- dict2doc ---------------------------------------------------------
    %
    function prompt = dict2doc(indent, name, format, meta)
      prompt = sprintf('\n%s%s', indent, format.kind);
      if name
        prompt = sprintf('%s %s', prompt, name);
      end
      if isfield(format, 'length')
        prompt = sprintf('%s[%d]', prompt, format.length);
      end
      if isfield(meta, 'doc')
        prompt = sprintf('%s: %s', prompt, meta.doc);
      end
      if ~strcmp(format.kind, 'optional')
        if isfield(meta, 'default')
          if isnumeric(meta.default)
            prompt = [prompt, ' (', num2str(meta.default), ')'];
          else
            prompt = [prompt, ' (', meta.default, ')'];
          end
        end
      end

      switch format.kind
        case {'struct', 'union'}
          if isfield(meta, 'items');
            items = meta.items;
          else
            items = struct();
          end
          for p = fieldnames(format.type)'
            if isfield(items, p); m = items.(p); else m = struct(); end
            prompt = [prompt, ...
                      genomix.client.dict2doc([indent, '| '], ...
                                              [name, '.', p{1}], ...
                                              format.type.(p{1}), m)];
          end

        case {'array', 'sequence'}
          if isfield(format, 'length')
            l = num2str(format.length);
          else
            l = '';
          end
          prompt = [prompt, ...
                    genomix.client.dict2doc([indent, '| '], ...
                                            [name, '[', l, ']'], ...
                                            format.type, struct())];

          if isfield(meta, 'items');
            items = meta.items;
            for i = 1:length(items)
              prompt = [prompt, ...
                        genomix.client.dict2doc([indent, '| | '], ...
                                                [name, '[',num2str(i),']'], ...
                                                format.type, items{i})];
            end
          end

        case 'optional'
          if isfield(meta, 'default'); m = meta.default; else m = struct(); end
          prompt = [prompt, ...
                    genomix.client.dict2doc([indent, '| '], ...
                                            '', format.type, m)];
      end
    end
  end
end

% matlab 2012 compat
function s = strjoin(args, sep)
  if isempty(args)
    s = '';
  else
    if nargin < 2; sep = ','; end
    s = [sprintf(['%s' sep], args{1:end-1}) args{end}];
  end
end
